#!/usr/bin/perl -w
#
# Files hosted by Stanford:
# - http://fah-web.stanford.edu/teamstats/team$TEAM_ID.txt
#   Per team stats
#   Updated every 3 hours
#
# - http://fah-web.stanford.edu/daily_user_summary.txt
#   User stats
#   Updated every 3 hours
#
# - http://fah-web.stanford.edu/daily_team_summary.txt
#   Team stats
#   Updated every 3 hours
#
# - http://fah-web.stanford.edu/psummary.html
#   Project information of all currently running general projects
#   Updated every hour
# 
# - http://fah-web.stanford.edu/serverstat.html
#   Information on the status of the project servers 
#   Updated every 15 minutes

use strict qw(subs vars);
use Cwd;
use File::Basename;
use File::Temp;
use Date::Calc;
use LWP::UserAgent;
use XML::Simple;
use Data::Dumper;
use Getopt::Mixed "nextOption";

# set the Date::Calc language
Date::Calc::Language(Date::Calc::Decode_Language('English'));

# show what's print()ed without waiting for the newline
$|=1;

my %cfg = ( self                       => 'fci-update-stanford-files.pl',
            version                    => '0.4.1',
	    
	    verbose                    => 0,
	    quiet                      => 0,
	    download                   => 1,
	    force                      => 0,
	    action                     => 'all',
	    
	    'www-data-dir'             => '/var/www/fci/',
	    'stanford-files-file'      => 'xml-data/stanford-files.xml',
	    'xml-project-summary-file' => 'xml-data/project-summary.xml',
	    'team-list-file'           => 'xml-data/team-list.xml',
	    'teamstats-list-file'      => 'xml-data/teamstats-list.xml',
	    'site-config-file'         => 'settings/site-data/config.xml',

	    'bunzip2-bin'              => '/bin/bunzip2',

            update_project_summary     => 0,
	    update_team_summary        => 0,
	    update_user_summary        => 0,
	    update_teamstats           => 0,
          );

# Get commandline parameters
Getopt::Mixed::init("dir=s
                     project-summary
                     team-summary
                     user-summary
		     team-stats
                     no-download
		     force
                     verbose
		     quiet
		     help
		     d>dir
		     p>project-summary
		     u>user-summary
		     t>team-summary
		     T>team-stats
		     n>no-download
		     f>force
		     v>verbose
		     q>quiet
		     h>help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit;
	}
	
	if($option eq 'project-summary') {
        	$cfg{update_project_summary} = 1;
		$cfg{action}                 = 'individual';
	}
	if($option eq 'team-summary') {
		$cfg{update_team_summary}    = 1;
		$cfg{action}                 = 'individual';
	}
	if($option eq 'user-summary') {
		$cfg{update_user_summary}    = 1;
		$cfg{action}                 = 'individual';
	}
	if($option eq 'team-stats') {
		$cfg{update_teamstats}       = 1;
		$cfg{action}                 = 'individual';
	}
	
	if($option eq 'force') {
		$cfg{force}    = 1;
	}
	if($option eq 'no-download') {
		$cfg{download} = 0;
	}
	
	$cfg{'www-data-dir'} = $value if($option eq 'dir');
	$cfg{verbose}        = 1      if($option eq 'verbose');
	$cfg{quiet}          = 1      if($option eq 'quiet');
}
Getopt::Mixed::cleanup();

# if force was requested, override no-download
if($cfg{force}) {
	$cfg{download} = 1;
}
# if quiet was requested together with verbose, disable verbose which contradicts quiet
if($cfg{quiet} && $cfg{verbose}) {
	#print "--quiet overrides --verbose\n";
	$cfg{verbose} = 0;
}
# add directory separator to the end of the www-data-dir path if missing
if($^O eq 'MSWin32') {
	# Windows
	$cfg{'www-data-dir'} .= '\\' if(substr($cfg{'www-data-dir'}, -1, 1) ne '\\');
}
else {
	# One of the UNIX flavors is assumed
	$cfg{'www-data-dir'} .= '/' if(substr($cfg{'www-data-dir'}, -1, 1) ne '/');
}

my $dir = Cwd::cwd;
if(-d $cfg{'www-data-dir'} && -x $cfg{'www-data-dir'}) {
	print "Changing working directory to: $cfg{'www-data-dir'}\n" if($cfg{verbose});
	Cwd::chdir $cfg{'www-data-dir'};
	print "\n" if($cfg{verbose});
}
else {
	print "Error: Cannot change to directory: $cfg{'www-data-dir'}\n";
	exit 1;
}


# check file existance and permissions
if(!-r $cfg{'team-list-file'}) {
	print "Error: Cannot read team-list XML file!\n";
	exit 1;
}
if(!-e $cfg{'teamstats-list-file'}) {
	open(F, ">", $cfg{'teamstats-list-file'}) || die "Error: Cannot open teamstats-list XML file ($!)";
	close F;
}
if(!-w $cfg{'teamstats-list-file'}) {
	print "Error: Cannot write teamstats-list XML file!\n";
	exit;
}

if(!-r $cfg{'stanford-files-file'}) {
	print "Error: Cannot read stanford-files XML file!\n";
	exit;
}
if(!-w $cfg{'stanford-files-file'}) {
	print "Error: Cannot write stanford-files XML file!\n";
	exit;
}

if(!-w $cfg{'xml-project-summary-file'}) {
	print "Error: Cannot write project-summary XML file!\n";
	exit;
}

# read the XML file files
my $config         = read_xml(%cfg, Name => 'site-config')    || die "Error: Cannot read site-config XML file!\n";
my $team_list      = read_xml(%cfg, Name => 'team-list')      || die "Error: Cannot read team-list XML file!\n";
my $teamstats_list = read_xml(%cfg, Name => 'teamstats-list') || die "Error: Cannot read teamstats-list XML file!\n";
my $stanford_files = read_xml(%cfg, Name => 'stanford-files') || die "Error: Cannot read stanford-files XML file!\n";

#print Dumper($stanford_files);
#print Dumper($config);

my @applications = qw(bunzip2);
foreach my $application (@applications) {
        # set the path of the binary in question
        if(exists $config->{settings}->{application_list}->{application}->{$application} &&
           exists $config->{settings}->{application_list}->{application}->{$application}->{path} &&
           -x $config->{settings}->{application_list}->{application}->{$application}->{path}
        ) {
                $cfg{$application.'-bin'} = $config->{settings}->{application_list}->{application}->{$application}->{path};
        }
        else {
                print "Error: Cannot find path to the $application binary!\n";
                exit 1;
        }
}


if($cfg{action} eq 'all') {
	$cfg{update_teamstats}       = 1;
	$cfg{update_team_summary}    = 1;
	$cfg{update_user_summary}    = 1;
	$cfg{update_project_summary} = 1;
}

if($cfg{update_teamstats}) {
	update_teamstats($team_list, $teamstats_list, %cfg);
	print "\n" if($cfg{verbose});
}
if($cfg{update_team_summary}) {
	update_stanford_file('team_summary', $stanford_files, %cfg);
	print "\n" if($cfg{verbose});
}
if($cfg{update_user_summary}) {
	update_stanford_file('user_summary', $stanford_files, %cfg);
	print "\n" if($cfg{verbose});
}
if($cfg{update_project_summary}) {
	my $update_projects = 0;

	my @files = qw(project_summary project_summary_beta project_summary_comprehensive);
	foreach my $file (@files) {
		if(exists $stanford_files->{stanford_files}->{file}->{$file} &&
		   exists $stanford_files->{stanford_files}->{file}->{$file}->{url}
		) {
			my $updated = 0;
			
			($updated, $stanford_files) = update_stanford_file($file, $stanford_files, %cfg);
			
			if($updated) {
				$update_projects = 1;
			}
			print "\n" if($cfg{verbose});
		}
	}
	
	if($update_projects) {
		update_projects($stanford_files, %cfg);
	}
}

################################[ Subroutines ]################################

sub get_date {
	my %option = @_;

	my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();
	
	$month = '0'.$month if(length($month) == 1);
	$day   = '0'.$day   if(length($day)   == 1);
	$hour  = '0'.$hour  if(length($hour)  == 1);
	$min   = '0'.$min   if(length($min)   == 1);
	$sec   = '0'.$sec   if(length($sec)   == 1);
	
	my $date = "$year-$month-$day";
	my $time = "$hour:$min:$sec";

	if(!$option{date} && !$option{time}) {
		return "$date $time";
	}
	elsif($option{time} && !$option{date}) {
		return $time;
	}
	elsif($option{date} && !$option{time}) {
		return $date;
	}
	else {
		return ($date, $time);
	}
}

##############
#
#     Routine: read_xml()
#  Parameters: %args
#     Returns: a hashreference of the XML structure
# Description: read an XML file while enforcing the correct structure
#

sub read_xml {
	my %args = @_;
	
	if(!$args{Name}) {
		warn "FCI read_xml: Unknown name!\n";	
		return undef;
	}

	if($args{Name} =~ /^(client|qd|team-stats|known-project)$/) {
		if(!$args{File}) {
			warn "FCI read_xml: No file specified! ($args{Name})\n";
			return undef;
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			warn "FCI read_xml: Unknown file: $args{Name}\n";
			return undef;	
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File}) {
		warn "FCI read_xml: File does not exist: $args{File}\n";
		return undef;
	}
	if(-e $args{File} && !-r $args{File}) {
		warn "FCI read_xml: Cannot read file: $args{File}\n";
		return undef;
	}
	if(!-f $args{File}) {
		warn "FCI read_xml: Not a file: $args{File}\n";
		return undef;
	}
	
	if($args{File} && -r $args{File} && -f $args{File}) {
		if($args{Name} eq 'qd-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' },
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			return XMLin( $args{File},
			              KeyAttr    => 'team_rank',
				      ForceArray => [ 'rank_entry' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team_stats => 'team' }, 
				      ForceArray => [ 'team_stats' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' }, 
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { project  => 'number', client => 'name' }, 
				      ForceArray => [ 'project', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			return XMLin( $args{File},
			              KeyAttr    => { image  => 'type',  setting  => 'id', 
				                      option => 'value', location => 'id' },
				      ForceArray => [ 'image', 'setting', 'option', 'location' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { project => 'number', image => 'type' },
				      ForceArray => [ 'project', 'image' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-project') {
			return XMLin( $args{File},
				      KeyAttr    => { trajectory => 'id', image => 'type', client => 'name' },
				      ForceArray => [ 'trajectory', 'image', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			return XMLin( $args{File},
			              KeyAttr    => { path => 'name' },
				      ForceArray => [ 'path' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-users') {
			return XMLin( $args{File},
			              KeyAttr    => { user => 'name', client => 'name' },
				      ForceArray => [ 'user', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-clients') {
			return XMLin( $args{File},
			              KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-access-control') {
			return XMLin( $args{File},
			              KeyAttr    => { acl => 'name', client => 'name', user => 'name' },
				      ForceArray => [ 'acl', 'client', 'user' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'qd') {
			return XMLin( $args{File},
				      KeyAttr    => { index => 'number', server => 'number' },
				      ForceArray => [ 'index', 'server' ]
				    ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name', 'server' => 'number' },
				      ForceArray => [ 'server' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team => 'number', username => 'userid' },
				      ForceArray => [ 'team', 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			return XMLin( $args{File},
			              KeyAttr    => { project => 'number' },
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'fci-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
                elsif($args{Name} eq 'jmol-missing-projects') {
                        return XMLin( $args{File},
                                      ForceArray => [ 'project' ],
                                      KeepRoot   => 1
                                    ) || die $@;
                }
		elsif($args{Name} eq 'jmol-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { fah_proj => 'number' },
				      ForceArray => [ 'fah_proj' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
                elsif($args{Name} eq 'queue-graphs') {
                        return XMLin( $args{File},
                                      ForceArray => [ 'client' ],
                                      KeepRoot   => 1
                                    ) || die $@;
                }
		else {
			# unknown file
			return undef;
		}
	}

	return undef;
}


##############
#
#     Routine: write_xml()
#  Parameters: name of the XML file 
#              xml structure in a hash reference
#              %args
#     Returns: a hashreference of the XML structure
#     Returns: 1 on sucess
#              0 and string with reasons on errors
# Description: write an XML file while enforcing the correct structure
#

sub write_xml {
	my %args = @_;
	
	my ($success, $reason) = (1, '');
	
	if(!$args{Name}) {
		return (0, "Unknown name!\n");
	}
	
	if($args{Name} =~ /^(client|qd|known-project)$/) {
		if(!$args{File}) {
			return (0, "No file specified! ($args{Name})\n");
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			return (0, "Unknown file: $args{Name}\n");
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File} && !-w dirname($args{File}) ) {
		return (0, "Cannot write directory for: $args{Name}\n");
	}
	if(-e $args{File} && !-w $args{File}) {
		return (0, "Cannot write file: $args{Name}\n");
	}
	if(-e $args{File} && !-f $args{File}) {
		return (0, "Not a file: ($args{Name})\n");
	}
	
	
	if($success) {
		if($args{Name} eq 'qd-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => 'team_rank',
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team_stats => 'team' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project  => 'number', client => 'name' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { image  => 'type',  setting  => 'id', 
				                option => 'value', location => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project => 'number', image => 'type' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'known-project') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { trajectory => 'id', image => 'type', client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { path => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-users') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { user => 'name', client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-clients') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-access-control') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { acl => 'name', client => 'name', user => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name', server => 'number', 'fci' => 'version' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team => 'number', username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { project => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-missing-projects') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-projects') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { fah_proj => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
                elsif($args{Name} eq 'queue-graphs') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		else {
			# unknown file
			$success = 0;
			$reason .= "Unknown file: $args{Name}\n";
		}
	}

	return ($success, $reason);
}

sub modified_url {
	my $url           = shift;
	my $last_modified = shift;
	my %cfg           = @_;

	my $ua  = LWP::UserAgent->new();
	   $ua->agent("$cfg{self}/$cfg{version}");
	
	print "Checking URL for modification: $url ... " if($cfg{verbose});
		
	my $req = HTTP::Request->new( HEAD => $url,
	                              HTTP::Headers->new('If-Modified-Since' => $last_modified)
	                            );
	my $res = $ua->request($req);
	if($res->code eq '304') {
		print "Not Modified.\n" if($cfg{verbose});
		
		return 0;
	}
	elsif($res->code >= 400 && $res->code < 600) {
		print "Failed!\n" if($cfg{verbose});
		
		print "Warning: Failed to check URL for modification: $url\n";
		print "Server returned: ".$res->code." ".$res->message."\n";

		return -1;
	}
	else {
		print "Modified!\n" if($cfg{verbose});

		return 1;
	}
}

sub update_teamstats {
	my $team_list      = shift;
	my $teamstats_list = shift;
	my %cfg            = @_;

	print "Updating Team Stats ... \n" if($cfg{verbose});
	
	my $updated = 0;
	my $teams   = 0;
	foreach my $team (keys %{$team_list->{team_list}->{team}}) {
		print "\nStats for team: $team\n" if($cfg{verbose});

		$teams++;
		
		my $teamstats_url   = 'http://fah-web.stanford.edu/teamstats/';
		if($teamstats_list->{team_stats_list}->{base_url}) {
			$teamstats_url   = $teamstats_list->{team_stats_list}->{base_url};
		}
		   $teamstats_url  .= 'team'.$team.'.txt';
		my $teamstats_file  = 'stanford-data/team'.$team.'.txt';
		my $xml_file        = 'xml-data/team-'.$team.'.xml';

		if($teamstats_list->{team_stats_list}->{team_stats}->{$team}->{last_modified}) {
			# check the teamstats file on the Stanford webserver
			# only continue for this team if it was modified since the last time we downloaded it.
		
			my $last_modified = $teamstats_list->{team_stats_list}->{team_stats}->{$team}->{last_modified};

			my $modified = modified_url($teamstats_url, $last_modified, %cfg);
			if(!$modified) {
				if($cfg{force}) {
					# we will only download files if they were modified
					# or if we're forced to
					
					print "Parameter --force given, downloading anyway.\n" if($cfg{verbose});
				}
				else {
					if($cfg{verbose}) {
						print "Team Stats for Team $team are up-to-date, skipping download.\n";
					}
					next;
				}
			}
			elsif($modified == -1) {
				# server returned 400 of 500 range error, skip download & processing
				if($cfg{verbose}) {
					print $teamstats_list->{teamstats}->{$team}->{name};
					print " teamstats failed last modification check, skipping download.\n";
				}

				exit 1;
			}
		}
		else {
			print "No modification data known\n" if($cfg{verbose});
		}
		
		my $file = $teamstats_file;
		if($file =~ s/^$cfg{'www-data-dir'}//) {}
		my $date = get_date();
	
		$teamstats_list->{team_stats_list}->{team_stats}->{$team}->{file}        = $file;
		$teamstats_list->{team_stats_list}->{team_stats}->{$team}->{name}        = $team_list->{team_list}->{team}->{$team}->{name};
		
		$teamstats_list->{team_stats_list}->{team_stats}->{$team}->{last_update} = $date;
		$teamstats_list->{team_stats_list}->{last_update}                        = $date;
	
		# The XML file is no longer used, remove it if it exists
		if(-e $xml_file && -w dirname($xml_file) ) {
			unlink($xml_file) || die "Error: Cannot remove XML teamstats file: $xml_file ($!)";
		
			if(exists $teamstats_list->{team_stats_list}->{team_stats}->{$team}->{xml_file}) {
				delete $teamstats_list->{team_stats_list}->{team_stats}->{$team}->{xml_file};
				$updated++;
			}
		}
		if(!-e $xml_file) {
			if(exists $teamstats_list->{team_stats_list}->{team_stats}->{$team}->{xml_file}) {
				delete $teamstats_list->{team_stats_list}->{team_stats}->{$team}->{xml_file};
				$updated++;
			}
		}

		# download team1.txt from Stanford
		if($cfg{download} == 1) {
			my $ua  = LWP::UserAgent->new();
			   $ua->agent("$cfg{self}/$cfg{version}");
			
			print "Downloading URL: $teamstats_url ... " if($cfg{verbose});
			
			my $req = HTTP::Request->new( GET => $teamstats_url );
			my $res = $ua->request($req);
			if($res->is_success) {
				print "Success\n" if($cfg{verbose});
	
				my $last_modified = $res->header('Last-Modified');
				print "Last-Modified: $last_modified\n" if($cfg{verbose});
	
				$teamstats_list->{team_stats_list}->{team_stats}->{$team}->{last_modified} = $last_modified;
				
				print "Saving file to disk: $teamstats_file\n" if($cfg{verbose});
				open(F, ">", $teamstats_file) || die "Cannot open file: $teamstats_file ($!)";
				print F $res->content;
				close F;

				print "Successfully updated the Team Stats for Team $team.\n" if($cfg{verbose});
				
				$updated++;
	
				print "\n" if($cfg{verbose});
			}
			else {
				print "Failed!\n" if($cfg{verbose});
				print "Error: Failed to retrieve URL! ($teamstats_url)\n";
				print "Server returned: ".$res->code." ".$res->message."\n";
				
				exit 1;
			}
		}
	}
	
	if($teams == 0) {
		print "No teams listed yet in the team-list XML file.\n";
	}
	
	if($updated > 0) {
		print "Saving teamstats-list XML file.\n" if($cfg{verbose});
        	my ($success, $reason) = write_xml( %cfg,
		                                    Name => 'teamstats-list',
		                                    XML  => $teamstats_list
		                                  );
		
        	if(!$success) {
        	        print "Error: $reason";
        	}
	}
}

sub bunzip_data {
	my $data = shift;
	my %cfg  = @_;

	# Save into temp file
	my $tmp = File::Temp->new(
				   DIR      => '/tmp/',
				   TEMPLATE => 'fci-update-stanford-files.XXXXXXXXXX',
				   SUFFIX   => '.bz2',
				   UNLINK   => 0,
				   OPEN     => 0,
				 );

	my $temp_file = $tmp->filename;

	open(F, '>', $temp_file) || die "Error: Cannot open temporary file: $temp_file ($!)";
	binmode F;
	print F $data;
	close F;

	# Decompress the temp file
	print "Decompressing with: $cfg{'bunzip2-bin'}\n" if($cfg{verbose});

	my @cmd  = ($cfg{'bunzip2-bin'}, $temp_file);
	my $exit = system(@cmd) >> 8;
	if($exit == 0) {
		(my $file = $temp_file) =~ s/.bz2$//;

		$data = '';
		open(F, '<', $file) || die "Error: Cannot open uncompressed file: $file ($!)";
		while(<F>) {
			$data .= $_;
		}
		close F;

		unlink($file) || die "Error: Cannot remove uncompressed file: $file ($!)";
	}
	else {
		print "Error: Failed to decompress: $temp_file\n";

		unlink($temp_file) || die "Error: Cannot remove temporary file: $temp_file ($!)";

		exit 1;
	}

	return $data;	
}

sub download_stanford_file {
	my $file           = shift;
	my $stanford_files = shift;
	my %cfg            = @_;

	if(!$stanford_files->{stanford_files}->{file}->{$file}) {
		print "Error: Stanford file not listed: $file\n";
		return;
	}

	if($cfg{verbose}) {
		print "Updating ";
		print $stanford_files->{stanford_files}->{file}->{$file}->{name};
		print " ... \n";
	}

	my $last_modified = '';
	my $content       = '';

	if($stanford_files->{stanford_files}->{file}->{$file}->{url}) {
		my $url            = $stanford_files->{stanford_files}->{file}->{$file}->{url};
		my $local_file     = $stanford_files->{stanford_files}->{file}->{$file}->{local_file};
		my $bz2_url        = $stanford_files->{stanford_files}->{file}->{$file}->{bz2_url}        ? $stanford_files->{stanford_files}->{file}->{$file}->{bz2_url}        : '';
		my $url_preference = $stanford_files->{stanford_files}->{file}->{$file}->{url_preference} ? $stanford_files->{stanford_files}->{file}->{$file}->{url_preference} : '';
		
		my $download_url = $url;
		   $download_url = $bz2_url if($bz2_url && $url_preference && $url_preference eq 'bz2_url');

		if($stanford_files->{stanford_files}->{file}->{$file}->{last_modified}) {
			# check the file on the Stanford webserver
			# only continue for this file if it was modified since the last time we downloaded it.

			my $last_modified = $stanford_files->{stanford_files}->{file}->{$file}->{last_modified};

			my $modified = modified_url($download_url, $last_modified, %cfg);
			if(!$modified) {
				# the file was not modified, skip download & processing
					
				if($cfg{force}) {
					# we will only download files if they were modified
					# or if we're forced to
					
					print "Parameter --force given, downloading anyway.\n" if($cfg{verbose});
				}
				else {
					if($cfg{verbose}) {
						print $stanford_files->{stanford_files}->{file}->{$file}->{name};
						print " is up-to-date, skipping download.\n";
					}
					
					$cfg{download} = 0;
				}
			}
			elsif($modified == -1) {
				# server returned 400 of 500 range error, skip download & processing
				if($cfg{verbose}) {
					print $stanford_files->{stanford_files}->{file}->{$file}->{name};
					print " failed last modification check, skipping download.\n";
				}

				exit 1;
			}
		}
		else {
			print "No modification data known\n" if($cfg{verbose});
		}
		
		
		# download file from Stanford
		if($cfg{download} == 1) {
			my $ua  = LWP::UserAgent->new();
			   $ua->agent("$cfg{self}/$cfg{version}");
			
			print "Downloading URL: $download_url ... " if($cfg{verbose});
			
			my $req = HTTP::Request->new( GET => $download_url );
			my $res = $ua->request($req);
			if($res->is_success) {
				print "Success\n" if($cfg{verbose});
	
				$last_modified = $res->header('Last-Modified');
				print "Last-Modified: $last_modified\n" if($cfg{verbose});

				$content = $res->content;

				if($download_url eq $bz2_url && $download_url =~ /\.bz2$/) {
					$content = bunzip_data($content, %cfg);
				}
			}
			else {
				print "Failed!\n" if($cfg{verbose});

				print "Error: Failed to retrieve URL: $download_url\n";
				print "Server returned: ".$res->code." ".$res->message."\n";
				
				exit 1;
			}
		}
	}

	return ($last_modified, $content);
}

sub update_stanford_file {
	my $file           = shift;
	my $stanford_files = shift;
	my %cfg            = @_;

	my ($last_modified, $content) = download_stanford_file($file, $stanford_files, %cfg);

	if($last_modified && $content) {
		my $local_file = $stanford_files->{stanford_files}->{file}->{$file}->{local_file};
	
		print "Saving file to disk: $local_file\n" if($cfg{verbose});
		open(F, ">", $local_file) || die "Cannot open file: $local_file ($!)";
		print F $content;
		close F;
	
		my $date = get_date();	

		$stanford_files->{stanford_files}->{file}->{$file}->{last_modified} = $last_modified;
		$stanford_files->{stanford_files}->{file}->{$file}->{last_update}   = $date;
	
		if($cfg{verbose}) {
			print "Successfully updated the ";
			print $stanford_files->{stanford_files}->{file}->{$file}->{name};
			print "\n";
		}
		
		print "Saving stanford-files XML file.\n" if($cfg{verbose});
	       	my ($success, $reason) = write_xml( %cfg,
		                                    Name => 'stanford-files',
		                                    XML  => $stanford_files
		                                  );
		
	       	if(!$success) {
	       	        print "Error: $reason";
	       	}

		return (1, $stanford_files);
	}

	return (0, $stanford_files);
}

sub update_projects {
	my $stanford_files = shift;
	my %cfg            = @_;

	print "Updating Project Summary XML file ... \n" if($cfg{verbose});
	
	my $project_summary = read_xml(%cfg, Name => 'xml-project-summary') || die "Error: Cannot read project-summary XML file!\n";
	
	# set the status for all projects to inactive
	# only the projects which are actually listed on the psummary*.html pages should be active
	foreach my $number (keys %{$project_summary->{project_summary}->{project}}) {
		# copy the current status for later comparison and removal
		if($project_summary->{project_summary}->{project}->{"$number"}->{status}->{active}) {
			$project_summary->{project_summary}->{project}->{"$number"}->{old_status}->{active} 
				= $project_summary->{project_summary}->{project}->{"$number"}->{status}->{active};
		}
		else {
			$project_summary->{project_summary}->{project}->{"$number"}->{old_status}->{active} 
				= 'false';
		}
		# reset the active flag
		$project_summary->{project_summary}->{project}->{$number}->{status}->{active} = 'false';

			
		# general flag
		if($project_summary->{project_summary}->{project}->{"$number"}->{status}->{general}) {
			$project_summary->{project_summary}->{project}->{"$number"}->{old_status}->{general} 
				= $project_summary->{project_summary}->{project}->{"$number"}->{status}->{general};
		}
		else {
			$project_summary->{project_summary}->{project}->{"$number"}->{old_status}->{general} 
				= 'false';
		}
		$project_summary->{project_summary}->{project}->{$number}->{status}->{general} = 'false';


		if(exists $stanford_files->{stanford_files}->{file}->{project_summary_beta} &&
		   exists $stanford_files->{stanford_files}->{file}->{project_summary_comprehensive}
		) {
			if($project_summary->{project_summary}->{project}->{"$number"}->{status}->{beta}) {
				$project_summary->{project_summary}->{project}->{"$number"}->{old_status}->{beta} 
					= $project_summary->{project_summary}->{project}->{"$number"}->{status}->{beta};
			}
			else {
				$project_summary->{project_summary}->{project}->{"$number"}->{old_status}->{beta} 
					= 'false';
			}
			$project_summary->{project_summary}->{project}->{$number}->{status}->{beta} = 'false';
	

			if($project_summary->{project_summary}->{project}->{"$number"}->{status}->{comprehensive}) {
				$project_summary->{project_summary}->{project}->{"$number"}->{old_status}->{comprehensive} 
					= $project_summary->{project_summary}->{project}->{"$number"}->{status}->{comprehensive};
			}
			else {
				$project_summary->{project_summary}->{project}->{"$number"}->{old_status}->{comprehensive} 
					= 'false';
			}
			$project_summary->{project_summary}->{project}->{$number}->{status}->{comprehensive} = 'false';
		}
	}

	#print Dumper($project_summary);

	my @psummary_files = qw(project_summary project_summary_beta project_summary_comprehensive);
	foreach my $file (@psummary_files) {
		if(exists $stanford_files->{stanford_files}->{file}->{$file} &&
		   $stanford_files->{stanford_files}->{file}->{$file}->{local_file}
		) {
			my $local_file = $stanford_files->{stanford_files}->{file}->{$file}->{local_file};
	
			# parse the HTML of psummary.html
			$project_summary = parse_psummary($file, $local_file, $project_summary, %cfg);
			#print Dumper($project_summary);
		}
	}
	
	# remove old_status 
	foreach my $number (keys %{$project_summary->{project_summary}->{project}}) {
		delete $project_summary->{project_summary}->{project}->{$number}->{old_status};
	}
	
	# update the projects XML file
	print "Updating XML file: $cfg{'xml-project-summary-file'}\n" if($cfg{verbose});
        my ($success, $reason) = write_xml( %cfg,
	                                    Name => 'xml-project-summary',
	                                    XML  => $project_summary
	                                  );

        if(!$success) {
                print "Error: $reason";
	}
	else {
		print "Successfully updated the Project Summary.\n" if($cfg{verbose});
	}
		
	# add xml file to stanford-files
	$stanford_files->{stanford_files}->{file}->{project_summary}->{xml_file} = $cfg{'xml-project-summary-file'};
		
	print "Saving stanford-files XML file.\n" if($cfg{verbose});
        ($success, $reason) = write_xml( %cfg,
	                                 Name => 'stanford-files',
	                                 XML  => $stanford_files
	                               );

        if(!$success) {
                print "Error: $reason";
        }
}

sub parse_psummary {
	my $type = shift;
	my $file = shift;
	my $xml  = shift;
	my %cfg  = @_;

	if(-r $file) {
		if($cfg{verbose}) {
			print "\n";
			print "Parsing: ";
			print basename($file);
			print " ...\n";
		}

		my %seen;
		my $new     = 0;
		my $updated = 0;
		my $total   = 0;
		my $line    = 0;
		my $date    = get_date();

		# parse the HTML of psummary.html
		# use to projects listed on it to update the master project list: project-summary.xml
		# set the general status of each project listed to true
		open (F, "<", $file) || die "Error: Cannot open file: $file ($!)";
		while (<F>) {
			$line++;

			# Columns per tablerow:
			#
			#  0 Project Number
			#  1 Server IP
			#  2 Work Unit Name
			#  3 Number of Atoms
			#  4 Preferred (days)
			#  5 Final Deadline (days)
			#  6 Credit
			#  7 Frames
			#  8 Code
			#  9 Description
			# 10 Contact
			# 11 Kfactor
			#	
			# example line to parse:
			#
			#<TR bgcolor="#dfdfdf"> <TD>6013</TD><TD> 130.237.232.140 </TD>
			#<TD>p6013_IBX in water</TD><TD>146808</TD><TD>    3.00</TD>
			#<TD>    3.00</TD><TD>  380.00</TD><TD>100</TD><TD>GRO-A3</TD>
			#<TD><a href=http://fah-web.stanford.edu/cgi-bin/fahproject?p=6013>Description</a></TD>
			#<TD><font size=-1>kasson</font></TD><TD>    3.36</TD></TR>
			#
			# alternate format of line to parse:
			#
			#<TR bgcolor="#dfdfdf"> <TD>5915</TD><TD> 171.64.65.20 </TD>
			#<TD>p5915_proL_ff96_300K</TD><TD>973</TD><TD>    3.00</TD>
			#<TD>    6.00</TD><TD> 1888.00</TD><TD>100</TD><TD>GROGPU2-MT</TD>
			#<TD><a href=http://fah-web.stanford.edu/cgi-bin/fahproject?p=5915>Description</a></TD>
			#<TD><font size=-1>ihaque</font></TD><TD><font size=-1>1</font></TD></TR>
			#
			# Yet another alternate format of line to parse:
			#
			#<TR bgcolor="#dfdfdf"> <TD>6013</TD><TD> 130.237.232.140 </TD>
			#<TD>p6013_IBX in water</TD><TD>146808</TD><TD>3.00</TD><TD>3.00</TD>
			#<TD>380.00</TD><TD>100</TD><TD>GRO-A3</TD>
			#<TD><a href="http://fah-web.stanford.edu/cgi-bin/fahproject?p=6013">Description</a></TD>
			#<TD><font size="-1">kasson</font></TD><TD>3.36</TD></TR>
			#
			# Another format with an email address as contact:
			#
			#<tr bgcolor="#ffffff"> <td>2963</td><td>143.89.28.70</td>
			#<td>p2963</td><td>2963</td><td>10.00</td><td>20.00</td>
			#<td>104.00</td><td>100</td><td>GROMACS</td>
			#<td><a href="http://fah-web.stanford.edu/cgi-bin/fahproject?p=2963">Description</a></td>
			#<td><font size="-1">xuhuihuang@gmail.com</font></td><td>0.00</td></tr>
			#
			#<tr bgcolor="#ffffff"> <td>10510</td><td>171.64.65.61</td>
			#<td>p10510_ntl9(1-56)_WT_300K</td><td>908</td><td>8.00</td><td>12.00</td>
			#<td>587.00</td><td>100</td><td>GROGPU2</td>
			#<td><a href="http://fah-web.stanford.edu/cgi-bin/fahproject.overusingIPswillbebanned?p=10510">Description</a></td>
			#<td><font size="-1">vvoelz</font></td><td>0.00</td></tr>
			#
			# There are multiple description URL formats in use:
			#
			# http://fah-web.stanford.edu/cgi-bin/fahproject?p=1487
			# http://fah-web.stanford.edu/cgi-bin/allprojects#1311
			# http://fah-web.stanford.edu/cgi-bin/fahproject.overusingIPswillbebanned?p=10103
			
			next unless /cgi-bin/;          # only parse project lines
			chomp;                          # remove start/end junk
		
			# Project Number | Server IP | Work Unit Name | Number of Atoms | Preferred (days) |
			# Final Deadline (days) | Credit | Frames | Code | Description | Contact | Kfactor
			if(m!^<TR bgcolor="#\w{6}">\s*<TD>\s*(\d+)\s*</TD><TD>\s*(\d+\.\d+\.\d+\.\d+)\s*</TD><TD>\s*([\w\-\+=\.,: \!\(\)]+?)\s*</TD><TD>\s*(\d+|)\s*</TD><TD>\s*(\d+\.\d+)\s*</TD><TD>\s*(\d+\.\d+)\s*</TD><TD>\s*(\d+\.\d+)\s*</TD><TD>\s*(\d+)\s*</TD><TD>\s*([\w\-]+)\s*</TD><TD><a href=(?:")?(http://fah-web\.stanford\.edu/cgi-bin/allprojects#\d+|http://fah-web\.stanford\.edu/cgi-bin/fahproject(?:\.overusingIPswillbebanned)?\?p=\d+)(?:")?>Description</a></TD><TD><font size=(?:")?-1(?:")?>([\w@\.]*?)</font></TD><TD>\s*(?:<font size=-1>)?(\d+|\d+\.\d+)(?:</font>)?\s*</TD></TR>\s*$!i) {
				my $project_number     = $1;
				my $server_ip          = $2;
				my $work_unit_name     = $3;
				my $number_of_atoms    = $4;
				my $preferred_deadline = $5;
				my $final_deadline     = $6;
				my $credit             = $7;
				my $frames             = $8;
				my $code               = $9;
				my $description        = $10;
				my $contact	       = $11;
				my $kfactor	       = $12;

				my @key   = ('server', 'name', 'atoms', 'preferred', 'deadline', 
				             'credit', 'frames', 'code', 'description', 'contact', 'kfactor');
				
				my @value = ($server_ip, $work_unit_name, $number_of_atoms, $preferred_deadline, $final_deadline, 
				             $credit, $frames, $code, $description, $contact, $kfactor);
				
				# update the master XML projects list for the current project
				if(!$xml->{project_summary}->{project}->{$project_number}) {
					print "     New project: $project_number\n" if(!$cfg{quiet});
					
					my $i = 0;
					foreach(@key) {
						if(!defined $value[$i]) {
							#print "Warning: Empty field: '$_' in project: $project_number\n";
							$value[$i] = 'n/a';
						}
						$xml->{project_summary}->{project}->{$project_number}->{$_} = $value[$i];
						$i++;
					}
					$xml->{project_summary}->{project}->{$project_number}->{last_update} = $date;
					
					$new++;
				}
				else {
					#print "Known project $project_number\n";
					
					my $i = 0;
					my $s = '';
		
					# check every field of for this project for changes against the XML file
					my $modified = 0;
					my @modification;
					foreach(@key) {
						if(!defined $value[$i] || $value[$i] =~ /^\s*$/) {
							#print "Warning: Empty field: '$_' in project: $project_number\n";
							$i++;
							next;
						}
						if($_ eq 'name' && $value[$i] =~ /^p${project_number}_$/) {
							#print "Warning: Empty project name '$value[$i]' in project: $project_number\n";
							$i++;
							next;
						}
						# Update project description to new format
						#
						# Old: http://fah-web.stanford.edu/cgi-bin/fahproject?p=2494
						# New: http://fah-web.stanford.edu/cgi-bin/fahproject.overusingIPswillbebanned?p=2494

						if($_ eq 'description' && $value[$i] =~ /^(.*?fahproject)(\?p=\d+)$/) {
							$value[$i] = $1 .'.overusingIPswillbebanned'. $2;							
						}
					
						if((!exists $xml->{project_summary}->{project}->{$project_number}->{$_} &&
						    $value[$i] !~ /^\s*$/
						   ) 
						   ||
						   (exists $xml->{project_summary}->{project}->{$project_number}->{$_}        &&
						    $xml->{project_summary}->{project}->{$project_number}->{$_} ne $value[$i] &&
						    $value[$i] !~ /^\s*$/
						   )
						) {
						   	$modified  = 1;

							my $old_value = 'n/a';
							if(exists $xml->{project_summary}->{project}->{$project_number}->{$_}) {
								$old_value = $xml->{project_summary}->{project}->{$project_number}->{$_};
							}
							
							my %mod = ( property  => $_,
							            old_value => $old_value,
								    new_value => $value[$i]
							          );
		
							push @modification, \%mod;
							
							# we're looping trough all the properties of a project
							# we only need to add it once to the updated projects counter
							if(!$seen{$project_number}) {
								$seen{$project_number} = 1;
								$updated++;
							}
							
							$xml->{project_summary}->{project}->{$project_number}->{$_}          = $value[$i];
							$xml->{project_summary}->{project}->{$project_number}->{last_update} = $date;
						}
						$i++;
					}
					if($modified) {
						if(!$cfg{quiet}) {
							print "Modified project: $project_number  ";
			
							my $i = 0;
							foreach my $mod (@modification) {
								if($i > 0) {
									my $l = length("Modified project: $project_number  ");
									print ' ' x $l;
								}
								print $mod->{property};
								print ": ";
								print $mod->{old_value};
								print " -> ";
								print $mod->{new_value};
								print "\n";
								
								$i++;
							}
						}
					}
				}
				$total++;
							
				my $old_status = 'unknown_status';
				my $new_status = 'true';

				if($xml->{project_summary}->{project}->{$project_number}->{old_status}->{active}) {
					$old_status = $xml->{project_summary}->{project}->{$project_number}->{old_status}->{active};
				}
	
				if($old_status ne $new_status) {
					print " Updated project: $project_number  Status active: $old_status -> $new_status\n" if($cfg{verbose});
				}
				
				# set the active status to true for the current project
				$xml->{project_summary}->{project}->{$project_number}->{status}->{active} = $new_status;


				# set the status for the project type of the current project
				if($type eq 'project_summary') {
					my $old_status = 'unknown_status';
					my $new_status = 'true';

					if($xml->{project_summary}->{project}->{$project_number}->{old_status}->{general}) {
						$old_status = $xml->{project_summary}->{project}->{$project_number}->{old_status}->{general};
					}
		
					if($old_status ne $new_status) {
						print " Updated project: $project_number  Status general: $old_status -> $new_status\n" if($cfg{verbose});
					}
				
					# set the general status to true for the current project
					$xml->{project_summary}->{project}->{$project_number}->{status}->{general} = $new_status;
				}
				elsif($type eq 'project_summary_beta') {
					my $old_status = 'unknown_status';
					my $new_status = 'true';

					if($xml->{project_summary}->{project}->{$project_number}->{old_status}->{beta}) {
						$old_status = $xml->{project_summary}->{project}->{$project_number}->{old_status}->{beta};
					}
		
					if($old_status ne $new_status) {
						print " Updated project: $project_number  Status beta: $old_status -> $new_status\n" if($cfg{verbose});
					}
				
					# set the beta status to true for the current project
					$xml->{project_summary}->{project}->{$project_number}->{status}->{beta} = $new_status;
				}
				elsif($type eq 'project_summary_comprehensive') {
					my $old_status = 'unknown_status';
					my $new_status = 'true';

					if($xml->{project_summary}->{project}->{$project_number}->{old_status}->{comprehensive}) {
						$old_status = $xml->{project_summary}->{project}->{$project_number}->{old_status}->{comprehensive};
					}
		
					if($old_status ne $new_status) {
						print " Updated project: $project_number  Status comprehensive: $old_status -> $new_status\n" if($cfg{verbose});
					}
				
					# set the comprehensive status to true for the current project
					$xml->{project_summary}->{project}->{$project_number}->{status}->{comprehensive} = $new_status;
				}
			}
			else {
				# Cannot parse line!
				print "Warning: Cannot parse project line (".basename($file).":$line):\n$_\n";
			}
		}

		# print update summary in verbose mode
		# or if there were new or updated projects even in non-verbose mode
		print "\n" if($cfg{verbose});
		if($cfg{verbose} == 1 || ($new > 0 || $updated > 0) && !$cfg{quiet}) {
			print "Project Summary Records:\n";
			print "Updated: $updated\n";
			print "    New: $new\n";
			print "  Total: $total\n";
		}
	}
	else {
		print "Error: Cannot read file: $file\n";
		exit 1;
	}
	
	return $xml;
}

sub print_usage {
	my %cfg = @_;
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --dir <DIR>        Path of the FCI server directory\n";
	print "-p, --project-summary  Download psummary.html (default: download all files)\n";
	print "-u, --user-summary     Download daily_user_summary.txt\n";
	print "-t, --team-summary     Download daily_team_summary.txt\n";
	print "-T, --team-stats       Download team\$ID.txt for the teams known to FCI\n";
	print "-n, --no-download      Don't download any files, use the ones on disk\n";
	print "-f, --force            Download the files even if they're already uptodate\n";
	print "-v, --verbose          Enable verbose output\n";
	print "-q, --quiet            Don't print any output at all, except for errors\n";
	print "-h, --help             Show this usage information\n";
}
